/* packet-map.c
 * Routines for MAP packet dissection
 *
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <glib.h>
#include <epan/packet.h>
#include <epan/asn1.h>
#include <epan/dissectors/packet-per.h>
#include <epan/prefs.h>

#include <stdio.h>
#include <string.h>

#include "packet-map.h"

#define PNAME  "MAP"
#define PSNAME "MAP"
#define PFNAME "map"
#define MAP_PORT 2003    /* BTP port */


void proto_reg_handoff_map(void);

/* Initialize the protocol and registered fields */
static int proto_map = -1;
static int global_map_port = MAP_PORT;

static int hf_map_MAP_PDU = -1;
#include "packet-map-hf.c"

/* Initialize the subtree pointers */
static int ett_map = -1;

#include "packet-map-ett.c"

#include "packet-map-fn.c"

guint32
dissect_per_UTF8String(tvbuff_t *tvb, guint32 offset, asn1_ctx_t *actx, proto_tree *tree, int hf_index, int min_len, int max_len, gboolean has_extension)
{
  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index, min_len, max_len, has_extension, NULL);
  return offset;
}

static void dissect_MAP_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_) {
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, FALSE, pinfo);
  dissect_map_MAP_PDU(tvb, 0, &asn1_ctx, tree, hf_map_MAP_PDU);
}

static void
dissect_map(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
    proto_item      *map_item = NULL;
    proto_tree      *map_tree = NULL;

    /* make entry in the Protocol column on summary display */
    col_set_str(pinfo->cinfo, COL_PROTOCOL, PNAME);

    /* Clear out stuff in the info column */
    col_clear(pinfo->cinfo,COL_INFO);

    /* create the map protocol tree */
    if (tree) {
        map_item = proto_tree_add_item(tree, proto_map, tvb, 0, -1, FALSE);
        map_tree = proto_item_add_subtree(map_item, ett_map);
        dissect_MAP_PDU(tvb, pinfo, map_tree);
    }
}
/*--- proto_register_map -------------------------------------------*/
void proto_register_map(void) {

  module_t *map_module;

  /* List of fields */
  static hf_register_info hf[] = {
    { &hf_map_MAP_PDU,
      { "MAP", "map.MAP_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},

#include "packet-map-hfarr.c"
  };

  /* List of subtrees */
  static gint *ett[] = {
                  &ett_map,
#include "packet-map-ettarr.c"
  };


  /* Register protocol */
  proto_map = proto_register_protocol(PNAME, PSNAME, PFNAME);
  register_dissector("map", dissect_map, proto_map);

  /* Register fields and subtrees */
  proto_register_field_array(proto_map, hf, array_length(hf));
  proto_register_subtree_array(ett, array_length(ett));

  /* Register preferences module */
  map_module = prefs_register_protocol(proto_map, proto_reg_handoff_map);

  /* Register a sample port preference   */
  prefs_register_uint_preference(map_module, "btp.dport", "BTP destination port",
				 "BTP destination port",
				 10, &global_map_port);

}


/*--- proto_reg_handoff_map ---------------------------------------*/
void
proto_reg_handoff_map(void)
{
  static gboolean initialized = FALSE;
  static dissector_handle_t map_handle;
  static guint16 map_port;

  if (!initialized) {
    map_handle = create_dissector_handle(dissect_map, proto_map);
    dissector_add_handle("btp.dport", map_handle);
    initialized = TRUE;
  } else {
    if (map_port != 0) {
      dissector_delete_uint("btp.dport", map_port, map_handle);
    }
  }
  if (global_map_port != 0) {
    dissector_add_uint("btp.dport", global_map_port, map_handle);
  }
  map_port = global_map_port;
}


