/* packet-spat.c
 * Routines for SPAT packet dissection
 *
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <glib.h>
#include <epan/packet.h>
#include <epan/asn1.h>
#include <epan/dissectors/packet-per.h>
#include <epan/prefs.h>

#include <stdio.h>
#include <string.h>

#include "packet-spat.h"

#define PNAME  "SPAT"
#define PSNAME "SPAT"
#define PFNAME "spat"
#define SPAT_PORT 2004    /* BTP port */


void proto_reg_handoff_spat(void);

/* Initialize the protocol and registered fields */
static int proto_spat = -1;
static int global_spat_port = SPAT_PORT;

static int hf_spat_SPAT_PDU = -1;
#include "packet-spat-hf.c"

/* Initialize the subtree pointers */
static int ett_spat = -1;

#include "packet-spat-ett.c"

#include "packet-spat-fn.c"

guint32
dissect_per_UTF8String(tvbuff_t *tvb, guint32 offset, asn1_ctx_t *actx, proto_tree *tree, int hf_index, int min_len, int max_len, gboolean has_extension)
{
  offset = dissect_per_octet_string(tvb, offset, actx, tree, hf_index, min_len, max_len, has_extension, NULL);
  return offset;
}

static void dissect_SPAT_PDU(tvbuff_t *tvb _U_, packet_info *pinfo _U_, proto_tree *tree _U_) {
  asn1_ctx_t asn1_ctx;
  asn1_ctx_init(&asn1_ctx, ASN1_ENC_PER, FALSE, pinfo);
  dissect_spat_SPAT_PDU(tvb, 0, &asn1_ctx, tree, hf_spat_SPAT_PDU);
}

static void
dissect_spat(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
    proto_item      *spat_item = NULL;
    proto_tree      *spat_tree = NULL;

    /* make entry in the Protocol column on summary display */
    col_set_str(pinfo->cinfo, COL_PROTOCOL, PNAME);

    /* Clear out stuff in the info column */
    col_clear(pinfo->cinfo,COL_INFO);

    /* create the spat protocol tree */
    if (tree) {
        spat_item = proto_tree_add_item(tree, proto_spat, tvb, 0, -1, FALSE);
        spat_tree = proto_item_add_subtree(spat_item, ett_spat);
        dissect_SPAT_PDU(tvb, pinfo, spat_tree);
    }
}
/*--- proto_register_spat -------------------------------------------*/
void proto_register_spat(void) {

  module_t *spat_module;

  /* List of fields */
  static hf_register_info hf[] = {
    { &hf_spat_SPAT_PDU,
      { "SPAT", "spat.SPAT_element",
        FT_NONE, BASE_NONE, NULL, 0,
        NULL, HFILL }},

#include "packet-spat-hfarr.c"
  };

  /* List of subtrees */
  static gint *ett[] = {
                  &ett_spat,
#include "packet-spat-ettarr.c"
  };


  /* Register protocol */
  proto_spat = proto_register_protocol(PNAME, PSNAME, PFNAME);
  register_dissector("spat", dissect_spat, proto_spat);

  /* Register fields and subtrees */
  proto_register_field_array(proto_spat, hf, array_length(hf));
  proto_register_subtree_array(ett, array_length(ett));

  /* Register preferences module */
  spat_module = prefs_register_protocol(proto_spat, proto_reg_handoff_spat);

  /* Register a sample port preference   */
  prefs_register_uint_preference(spat_module, "btp.dport", "BTP destination port",
				 "BTP destination port",
				 10, &global_spat_port);

}


/*--- proto_reg_handoff_spat ---------------------------------------*/
void
proto_reg_handoff_spat(void)
{
  static gboolean initialized = FALSE;
  static dissector_handle_t spat_handle;
  static guint16 spat_port;

  if (!initialized) {
    spat_handle = create_dissector_handle(dissect_spat, proto_spat);
    dissector_add_handle("btp.dport", spat_handle);
    initialized = TRUE;
  } else {
    if (spat_port != 0) {
      dissector_delete_uint("btp.dport", spat_port, spat_handle);
    }
  }
  if (global_spat_port != 0) {
    dissector_add_uint("btp.dport", global_spat_port, spat_handle);
  }
  spat_port = global_spat_port;
}


